@@ -85,9 +85,8 @@ from .functionfs import (
     BIND, UNBIND, ENABLE, DISABLE, SETUP, SUSPEND, RESUME,
 )
 # pylint: enable=no-name-in-module
-from ._version import get_versions
-__version__ = get_versions()['version']
-del get_versions
+from . import _version
+__version__ = _version.get_versions()['version']
 
 __all__ = (
     'ch9',
@@ -172,7 +171,7 @@ def getInterfaceInAllSpeeds(interface, endpoint_list, class_descriptor_list=()):
           getDescriptor(USBSSPIsocEndpointDescriptor, ...)
         Must be provided and non-empty only when endpoint is isochronous and
         "superspeed" dict has "bmAttributes" bit 7 set.
-    class_descriptor (list of descriptors of any type)
+    class_descriptor_list (list of descriptors of any type)
       Descriptors to insert in all speeds between the interface descriptor and
       endpoint descriptors.
 
@@ -375,32 +374,6 @@ def getOSExtPropDesc(data_type, name, value):
         bProperty=value,
     )
 
-#def getDescs(*args, **kw):
-#    """
-#    Return a legacy format FunctionFS suitable for serialisation.
-#    Deprecated as of 3.14 .
-#
-#    NOT IMPLEMENTED
-#    """
-#    warnings.warn(
-#        DeprecationWarning,
-#        'Legacy format, deprecated as of 3.14.',
-#    )
-#    raise NotImplementedError('TODO')
-#    klass = type(
-#        'Descs',
-#        (DescsHead, ),
-#        {
-#            'fs_descrs': None, # TODO
-#            'hs_descrs': None, # TODO
-#        },
-#    )
-#    return klass(
-#        magic=DESCRIPTORS_MAGIC,
-#        length=ctypes.sizeof(klass),
-#        **kw
-#    )
-
 def getDescsV2(flags, fs_list=(), hs_list=(), ss_list=(), os_list=(), eventfd=None):
     """
     Return a FunctionFS descriptor suitable for serialisation.
@@ -832,18 +805,21 @@ class EndpointOUTFile(EndpointFile):
     """
     Read-only endpoint file.
     """
-    def __init__(self, path, submit, aio_block_list):
+    def __init__(self, path, submit, release, aio_block_list):
         """
         path (string)
             Endpoint file path.
         submit (AIOContext.submit)
             To submit AIOBlocks to after completion.
+        release ((AIOBlock) -> None)
+            Called when a comleted AIOBlock is not resubmitted.
         aio_block_list (list of AIOBlock)
             Blocks which belong to this endpoint. Modified to bind them to
             this file object (target_file and onCompletion).
         """
         super().__init__(path)
         self._submit = submit
+        self._release = release
         for aio_block in aio_block_list:
             aio_block.target_file = self
             aio_block.onCompletion = self._onComplete
@@ -880,7 +856,9 @@ class EndpointOUTFile(EndpointFile):
             data=data,
             status=status,
         )
-        if res != -errno.ESHUTDOWN:
+        if res == -errno.ESHUTDOWN:
+            self._release(aio_block)
+        else:
             # XXX: is it good to resubmit on any other error ?
             # XXX: what should be done on EAGAIN ?
             self._submit((aio_block, ))
@@ -900,17 +878,36 @@ class EndpointOUTFile(EndpointFile):
         May be overridden in subclass.
         """
 
+# FunctionFS can queue up to 4 events, so let's read that much.
+_EP0_EVENT_LIST_TYPE = Event * 4
+_EP0_EVENT_SIZE = ctypes.sizeof(Event)
+
+# XXX: how reliable is kernel version checking ?
+_, _, _KERNEL_VERSION, _, _ = os.uname()
+
 class Function:
     """
     Pythonic class for interfacing with FunctionFS.
 
-    Properties available:
+    Instance properties available:
     function_remote_wakeup_capable (bool)
         Whether the function wishes to be allowed to wake host.
     function_remote_wakeup (bool)
         Whether host has allowed the function to wake it up.
         Set and cleared by onSetup by calling enableRemoteWakeup and
         disableRemoteWakeup, respectively.
+    all_ctrl_recip, config0_setup:
+      Same value as given to constructor.
+      For forward compatibility, do not modify these properties (currently it
+      has no effect on the function at all).
+
+    Class properties available:
+    quirks_ffs_unsafe_eventfd (bool)
+        Whether the current kernel has issues with f_fs eventfd support.
+        Initialised on module intialisation with a simple condition on
+        kernel-version. May be changed prior to instanciation to override
+        this heuristic either way. Modification does not affect existing
+        instances.
     """
     _open = False
     _ep_list = ()
@@ -919,6 +916,17 @@ class Function:
     function_remote_wakeup_capable = False
     function_remote_wakeup = False
 
+    # f_fs, before kernel version 5.16, had a bug which caused the
+    # kernel-internal eventfd reference counter to underflow during gadget
+    # teardown. Depending on the exact version, maybe build options and
+    # maybe archivtecture, this could lead to kernel panics.
+    # Add backward-compatibility with such versions by working around the
+    # issue: do not give an eventfd to f_fs, and instead submit an extra AIO
+    # operation to still get eventfd notified when ep0 becomes readable.
+    quirks_ffs_unsafe_eventfd = (
+        _KERNEL_VERSION < '5.16'
+    )
+
     def __init__(
         self,
         path,
@@ -952,7 +960,7 @@ class Function:
             any configuration gets enabled.
         in_aio_blocks_max (int)
             Maximum number of IN transfers in-flight at any given time.
-            Submitting more transfers (using submitIN) will raise
+            Submitting more transfers (using submitIN) may raise
             OSError(EAGAIN).
             There may be a system-wide limit (64k AIO transfers as of 4.18).
         out_aio_blocks_per_endpoint (int)
@@ -968,18 +976,27 @@ class Function:
         self._path = path
         self._ep_address_dict = ep_address_dict = {}
         self._eventfd = eventfd = libaio.EventFD(flags=libaio.EFD_NONBLOCK)
+        self.all_ctrl_recip = all_ctrl_recip
+        self.config0_setup = config0_setup
         flags = 0
         if all_ctrl_recip:
             flags |= ALL_CTRL_RECIP
         if config0_setup:
             flags |= CONFIG0_SETUP
+        self.__quirks_ffs_unsafe_eventfd = quirks_ffs_unsafe_eventfd = (
+            self.quirks_ffs_unsafe_eventfd
+        )
         self._function_descriptor = getDescsV2(
             flags,
             fs_list=fs_list,
             hs_list=hs_list,
             ss_list=ss_list,
             os_list=os_list,
-            eventfd=eventfd,
+            eventfd=(
+                None
+                if quirks_ffs_unsafe_eventfd else
+                eventfd
+            ),
         )
         self._function_strings = getStrings(dict(lang_dict))
         self._out_aio_block_list = out_aio_block_list = []
@@ -1000,7 +1017,8 @@ class Function:
                     for _ in range(out_aio_blocks_per_endpoint):
                         # Using mmap to get a page-aligned buffer. f_fs strongly
                         # recommends aligning IN buffers to wMaxPacketSize
-                        # addresses, as this may be required by some UDCs.
+                        # addresses, as this may be required by some UDCs,
+                        # assume the same applies to OUT endpoints.
                         # Assume wMaxPacketSize will be less than a page.
                         out_block = libaio.AIOBlock(
                             mode=libaio.AIOBLOCK_MODE_READ,
@@ -1016,6 +1034,20 @@ class Function:
                         )
                         ep_aio_block_list.append(out_block)
                         out_aio_block_list.append(out_block)
+        self._ep0_event_list = _EP0_EVENT_LIST_TYPE()
+        if quirks_ffs_unsafe_eventfd:
+            # Piggy-back on the "in" AIO context for ep0 poll AIO block
+            in_aio_blocks_max += 1
+            self.__ep0_aio_block = libaio.AIOBlock(
+                mode=libaio.AIOBLOCK_MODE_POLL,
+                eventfd=eventfd,
+                onCompletion=self.__onEndpoint0AIOCompletion,
+                event_mask=select.EPOLLIN,
+            )
+            # Disable EP0 event processing in processEvents.
+            self.__processEndpoint0Events = lambda: None
+        else:
+            self.__processEndpoint0Events = self.__real_processEndpoint0Events
         if out_aio_block_list:
             self._out_aio_context = libaio.AIOContext(
                 len(out_aio_block_list),
@@ -1023,10 +1055,6 @@ class Function:
         self._in_aio_context = libaio.AIOContext(
             in_aio_blocks_max,
         )
-        # FunctionFS can queue up to 4 events, so let's read that much.
-        self._ep0_event_array_type = ep0_event_array_type = Event * 4
-        self._ep0_event_size = ctypes.sizeof(Event)
-        self._ep0_event_array_size = ctypes.sizeof(ep0_event_array_type)
 
     def __enter__(self):
         """
@@ -1048,6 +1076,7 @@ class Function:
                     is_in=is_in,
                     descriptor=descriptor,
                 )
+                # pylint: disable=unexpected-keyword-arg,no-value-for-parameter
                 if is_in:
                     ep_file = endpoint_class(
                         path=endpoint_path,
@@ -1058,14 +1087,19 @@ class Function:
                     ep_file = endpoint_class(
                         path=endpoint_path,
                         submit=self._out_aio_context.submit,
+                        release=self._out_aio_block_list.append,
                         aio_block_list=out_aio_block_dict[index],
                     )
+                # pylint: enable=unexpected-keyword-arg,no-value-for-parameter
                 ep_list.append(ep_file)
             fcntl.fcntl(
                 ep0,
                 fcntl.F_SETFL,
                 fcntl.fcntl(ep0, fcntl.F_GETFL) | os.O_NONBLOCK,
             )
+            if self.__quirks_ffs_unsafe_eventfd:
+                self.__ep0_aio_block.target_file = ep0
+                self._in_aio_context.submit((self.__ep0_aio_block, ))
         except:
             self.__unenter()
             raise
@@ -1076,11 +1110,12 @@ class Function:
         """
         Undo what __enter__ did.
         """
+        if self.__quirks_ffs_unsafe_eventfd:
+            self.__ep0_aio_block.target_file = None
         self._in_aio_context.cancelAll()
         out_aio_context = self._out_aio_context
         if out_aio_context is not None:
             out_aio_context.cancelAll()
-            self._out_aio_block_list = []
         ep_list = self._ep_list
         while ep_list:
             ep_list.pop().close()
@@ -1146,10 +1181,7 @@ class Function:
         """
         Process events until either an exception occurs or close is called.
         """
-        # Intent: "with select.epoll(1) as epoll:"
-        # But it only became a context manager in python3...
-        epoll = select.epoll(1)
-        try:
+        with select.epoll(1) as epoll:
             epoll.register(self.eventfd, select.EPOLLIN)
             poll = epoll.poll
             processEvents = self.processEvents
@@ -1161,15 +1193,13 @@ class Function:
                         raise
                 else:
                     processEvents()
-        finally:
-            epoll.close()
 
     def processEvents(self):
         """
         Process any available event (both functionfs events and
         AIO completions).
 
-        Non-blocking. Should be called whenever ep0 or eventfd become readable.
+        Non-blocking. Should be called whenever eventfd become readable.
         """
         try:
             # Rearm before calling getEvent, so that we do not risk skipping
@@ -1194,15 +1224,29 @@ class Function:
             out_aio_context = self._out_aio_context
             if out_aio_context is not None:
                 out_aio_context.getEvents(0)
-        buf = bytearray(self._ep0_event_array_size)
-        length = self.ep0.readinto(buf)
+        self.__processEndpoint0Events()
+
+    def __onEndpoint0AIOCompletion(self, block, res, _):
+        """
+        AIO block completion callback when quirks_ffs_unsafe_eventfd is true.
+        """
+        self.__real_processEndpoint0Events()
+        if res != -errno.ESHUTDOWN:
+            self._in_aio_context.submit((block, ))
+
+    def __real_processEndpoint0Events(self):
+        """
+        Read and handle endpoint 0 kernel events.
+        """
+        event_list = self._ep0_event_list
+        length = self.ep0.readinto(event_list)
         if length:
             event_dict = self.__event_dict
-            count, remainder = divmod(length, self._ep0_event_size)
-            assert remainder == 0, (length, self._ep0_event_size)
-            event_list = self._ep0_event_array_type.from_buffer(buf)
-            for index in range(count):
+            index = 0
+            while length >= _EP0_EVENT_SIZE:
                 event = event_list[index]
+                index += 1
+                length -= _EP0_EVENT_SIZE
                 event_type = event.type
                 if event_type == SETUP:
                     setup = event.u.setup
@@ -1219,7 +1263,8 @@ class Function:
                         self.ep0.halt(setup.bRequestType)
                         raise
                 else:
-                    getattr(self, event_dict[event.type])()
+                    getattr(self, event_dict[event_type])()
+            assert length == 0, (index, length, _EP0_EVENT_SIZE)
 
     def getEndpoint(self, index):
         """
@@ -1253,13 +1298,14 @@ class Function:
         Called when FunctionFS signals the function was (re)enabled.
         This may happen several times without onDisable being called.
         It must reset the function to its default state.
-        Also, submits transfers to all IN endpoints (if any).
+        Also, submits transfers to all OUT endpoints (if any).
 
         May be overridden in subclass.
         """
         self.disableRemoteWakeup()
         if self._out_aio_block_list:
             self._out_aio_context.submit(self._out_aio_block_list)
+            del self._out_aio_block_list[:]
 
     def onDisable(self):
         """
