@@ -80,6 +80,7 @@ class Gadget:
     """
     udb_gadget_path = '/sys/kernel/config/usb_gadget/'
     class_udc_path = '/sys/class/udc/'
+    __function_list = ()
 
     def __init__(
         self,
@@ -90,7 +91,7 @@ class Gadget:
         bcdDevice=None,
         bcdUSB=None,
         bDeviceClass=None,
-        bDeviceSubclass=None,
+        bDeviceSubClass=None,
         bDeviceProtocol=None,
         name=None,
         udc=None,
@@ -119,7 +120,7 @@ class Gadget:
         bcdUSB (int, None)
         bDeviceProtocol (int, None)
         bDeviceClass (int, None)
-        bDeviceSubclass (int, None)
+        bDeviceSubClass (int, None)
             See the USB specification for device descriptors.
             If None, the kernel default will be used.
             Some of these default values may prevent the gadget from working
@@ -162,10 +163,10 @@ class Gadget:
         elif not os.path.exists(os.path.join(self.class_udc_path, udc)):
             raise ValueError('No such UDC')
         self.__udc = udc
-        self.__config_list = list(enumerate(
+        self.__config_list = list(
             (
                 {
-                    'function_list': tuple(enumerate(config_dict['function_list'])),
+                    'function_list': tuple(config_dict['function_list']),
                     'attribute_dict': {
                         attribute_name: cast(
                             config_dict[attribute_name],
@@ -189,8 +190,7 @@ class Gadget:
                 }
                 for config_dict in config_list
             ),
-            1,
-        ))
+        )
         self.__lang_dict = {
             hex(lang): {
                 message_name: message_dict[message_name].encode('utf-8')
@@ -212,7 +212,7 @@ class Gadget:
                 'bcdUSB': bcdUSB,
                 'bDeviceProtocol': bDeviceProtocol,
                 'bDeviceClass': bDeviceClass,
-                'bDeviceSubclass': bDeviceSubclass,
+                'bDeviceSubClass': bDeviceSubClass,
             }.items()
             if value is not None
         }
@@ -227,7 +227,6 @@ class Gadget:
         )
         self.__name = name
         self.__real_name = None # chosen on __enter__
-        self.__function_list = []
         self.__dir_list = []
         self.__link_list = []
         self.__udc_path = None
@@ -301,12 +300,15 @@ class Gadget:
                 desc_file.write(value)
         dir_list.extend(self.__writeLangDict(name, self.__lang_dict))
         self.__writeAttributeDict(name, self.__attribute_dict)
-        function_list = self.__function_list
+        function_list = self.__function_list = []
         function_number_iterator = itertools.count()
         name_set = set()
         functions_root = os.path.join(name, 'functions')
         configs_root = os.path.join(name, 'configs')
-        for configuration_index, configuration_dict in self.__config_list:
+        for configuration_index, configuration_dict in enumerate(
+            self.__config_list,
+            1,
+        ):
             config_path = os.path.join(configs_root, 'c.%i' % (configuration_index, ))
             mkdir(config_path)
             dir_list.extend(
@@ -319,7 +321,9 @@ class Gadget:
                 config_path,
                 configuration_dict['attribute_dict'],
             )
-            for function_index, function in configuration_dict['function_list']:
+            for function_index, function in enumerate(
+                configuration_dict['function_list'],
+            ):
                 function_name = function.name
                 if function_name is None:
                     while True:
@@ -352,14 +356,15 @@ class Gadget:
                 )
         for function in function_list:
             function.wait()
-        self.__udc_path = udc_path = os.path.join(name, 'UDC')
+        udc_path = os.path.join(name, 'UDC')
         try:
             with open(udc_path, 'w') as udc:
                 udc.write(self.__udc)
         except (IOError, OSError) as exc:
             if exc.errno == 524: # ENOTSUPP, which is not ENOTSUP
                 exc.strerror = 'UDC cannot allocate this many endpoints'
-            raise
+            raise exc from None
+        self.__udc_path = udc_path
 
     def __exit__(self, exc_type, exc_value, tb):
         self.__unenter()
@@ -392,8 +397,7 @@ class Gadget:
                     file=sys.stderr,
                 )
                 traceback.print_exc()
-        while function_list:
-            function = function_list.pop()
+        for function in function_list:
             try:
                 function.join()
             except Exception: # pylint: disable=broad-except
@@ -426,6 +430,37 @@ class Gadget:
                 )
         self.__real_name = None
 
+    def getFunction(self, configuration_index, function_index):
+        """
+        Retrieve a function instance.
+        Only valid after __enter__ has been called once.
+
+        configuration_index (int)
+            Index of the configuration whose function is to be retrieved.
+        function_index (int)
+            Index, in the configuration, of the function to retrieve.
+
+        Note: arguments are both 0-based indexes, unlike in the USB protocol
+        where configurations indexes are 1-based.
+        """
+        config_list = self.__config_list
+        if (
+            configuration_index < 0 or
+            not 0 <= function_index < len(
+                config_list[configuration_index]['function_list']
+            )
+        ):
+            raise IndexError
+        for index in range(configuration_index):
+            function_index += len(config_list[index]['function_list'])
+        return self.__function_list[function_index]
+
+    def _iterFunctions(self):
+        """
+        Iterate over function instances.
+        """
+        return iter(self.__function_list)
+
 class _UsernameAction(argparse.Action):
     def __call__(self, parser, namespace, values, option_string=None):
         passwd = pwd.getpwnam(values)
@@ -505,9 +540,37 @@ class GadgetSubprocessManager(Gadget):
             **kw
         )
 
+    def _raiseKeyboardInterruptIfFunctionExited(
+        self,
+        signal_number,
+        stack_frame,
+    ):
+        """
+        Intended as SIGCHLD signal handler.
+        Raise KeyboardInterrupt if any of our function exited.
+        Allows the function implementation to spawn other processes and these
+        processes exiting without causing the gadget to wind down.
+        """
+        _ = signal_number # Silence pylint.
+        _ = stack_frame # Silence pylint.
+        if any(
+            x.getExitStatus() is not None
+            for x in self._iterFunctions()
+        ):
+            signal.signal(signal.SIGCHLD, signal.SIG_DFL)
+            raise KeyboardInterrupt
+
     def __enter__(self):
         super().__enter__()
-        signal.signal(signal.SIGCHLD, _raiseKeyboardInterrupt)
+        # Note: it is important to set SIGCHLD handler after calling
+        # super().__enter__(), as otherwise the subprocess would get such
+        # sighandler.
+        # XXX: Should ConfigFunctionFFSSubprocess.start reset all sighandlers
+        # to SIG_DFL ? This seems like a lot of work for (usually) no benefits.
+        signal.signal(
+            signal.SIGCHLD,
+            self._raiseKeyboardInterruptIfFunctionExited,
+        )
         # We are on the same terminal as subprocesses, so we will be getting
         # SIGINT at the same time as them. But we need to wait for them to
         # cleanup and then we will be notified by SIGCHLD.
@@ -572,7 +635,6 @@ class ConfigFunctionBase:
         Useful mostly if "start" method spawned processes/threads which need
         to do further initialisation on their own.
         """
-        raise NotImplementedError
 
     def kill(self):
         """
@@ -584,7 +646,6 @@ class ConfigFunctionBase:
         functions which do not spawn anything) so such functions can wind down
         in parallel.
         """
-        raise NotImplementedError
 
     def join(self):
         """
@@ -594,47 +655,75 @@ class ConfigFunctionBase:
         """
         raise NotImplementedError
 
-class ConfigFunctionKernel(ConfigFunctionBase):
+    def getExitStatus(self): # pylint: disable=no-self-use
+        """
+        Return the integer (typically 0 for success, and 1..127 for errors) exit
+        status of this function.
+
+        The meaning of this status if function-implementation-dependent.
+        Returns None when the function has not exited, or if there is no
+        meaningful status to collect for such function (ex: in-kernel
+        functions).
+        """
+        return None
+
+class ConfigFunctionKernel(
+    ConfigFunctionBase,
+): # pylint: disable=abstract-method
     """
     Base class for config functions which are implemented in the kernel.
     """
-    def __init__(self, config_dict=(), name=None):
+    __path = None
+
+    def __init__(self, config_dict=(), name=None, uid=None, gid=None):
         """
         config_dict (dict)
             key (str): path, relative to the function, of the option to set.
             value (str): value of the option
+        uid, gid (any)
+            Ignored. For compatibility with GadgetSubprocessManager.
         """
+        _ = uid # Silence pylint.
+        _ = gid # Silence pylint.
         self.__config_dict = dict(config_dict)
         super().__init__(name=name)
 
+    def _getOptionAbsPath(self, path, option_path):
+        """
+        Check wether option is a valid name (does not check if the option
+        exists).
+        """
+        self.__path = path
+        option_abspath = os.path.normpath(os.path.join(path, option_path))
+        if os.path.commonprefix((path, option_abspath)) != path:
+            raise ValueError('Invalid option path: %r' % (option_path, ))
+        return option_abspath
+
     def start(self, path):
         """
         Apply the content of config_dict.
         """
         for option_path, option_value in self.__config_dict.items():
-            option_abspath = os.path.normpath(os.path.join(path, option_path))
-            if os.path.commonprefix((path, option_abspath)) != path:
-                raise ValueError('Invalid option path: %r' % (option_path, ))
-            with open(option_abspath, 'w') as option_file:
+            with open(
+                self._getOptionAbsPath(path, option_path),
+                'w',
+            ) as option_file:
                 option_file.write(option_value)
 
-    def wait(self):
-        """
-        No-op.
-        """
-        return
-
-    def kill(self):
-        """
-        No-op.
-        """
-        return
-
     def join(self):
         """
         No-op.
         """
-        return
+
+    def getOption(self, option_path):
+        """
+        Read current option value.
+        """
+        with open(
+            self._getOptionAbsPath(self.__path, option_path),
+            'r',
+        ) as option_file:
+            return option_file.read()
 
 class ConfigFunctionFFS(ConfigFunctionBase): # pylint: disable=abstract-method
     """
@@ -642,6 +731,7 @@ class ConfigFunctionFFS(ConfigFunctionBase): # pylint: disable=abstract-method
     """
     type_name = "ffs"
     _mountpoint = None
+    function = None
 
     def __init__(
         self,
@@ -695,7 +785,7 @@ class ConfigFunctionFFS(ConfigFunctionBase): # pylint: disable=abstract-method
         """
         return self._getFunction(path=path)
 
-    def start(self, path):
+    def _mount(self, path):
         """
         Mount functionfs and set _mountpoint.
         """
@@ -727,7 +817,14 @@ class ConfigFunctionFFS(ConfigFunctionBase): # pylint: disable=abstract-method
         )
         self._mountpoint = mountpoint
 
-    def join(self):
+    def start(self, path):
+        self._mount(path)
+        self.function = function = self.getFunction(
+            path=self._mountpoint,
+        )
+        function.__enter__()
+
+    def _umount(self):
         """
         Unmount functionfs and clear _mountpoint.
         """
@@ -741,6 +838,10 @@ class ConfigFunctionFFS(ConfigFunctionBase): # pylint: disable=abstract-method
         os.rmdir(mountpoint)
         self._mountpoint = None
 
+    def join(self):
+        self.function.__exit__(None, None, None)
+        self._umount()
+
 class ConfigFunctionFFSSubprocess(ConfigFunctionFFS):
     """
     Function is isolated in a subprocess, allowing a change of user and group.
@@ -748,7 +849,7 @@ class ConfigFunctionFFSSubprocess(ConfigFunctionFFS):
     well-behaved service), beware of relative paths !
     """
     __pid = None
-    function = None
+    __exit_status = None
 
     def __init__(self, *args, **kw):
         super().__init__(*args, **kw)
@@ -763,11 +864,11 @@ class ConfigFunctionFFSSubprocess(ConfigFunctionFFS):
         In the parent process: return the callables expected by Gadget.
         """
         signal.signal(signal.SIGTERM, _raiseKeyboardInterrupt)
-        super().start(path)
+        self._mount(path)
         self.__pid = pid = os.fork()
         if pid == 0:
             try:
-                status = os.EX_OK
+                status = 0
                 os.close(self.__read_pipe)
                 self.__read_pipe = None
                 os.chdir('/')
@@ -781,9 +882,15 @@ class ConfigFunctionFFSSubprocess(ConfigFunctionFFS):
                 with function:
                     os.write(self.__write_pipe, _READY_MARKER)
                     self.run()
+            except SystemExit as exc:
+                if exc.code is not None:
+                    status = exc.code
+                    if not isinstance(status, int):
+                        print(status, file=sys.stderr)
+                        status = 1
             except: # pylint: disable=bare-except
                 traceback.print_exc()
-                status = os.EX_SOFTWARE
+                status = 1
             finally:
                 os.close(self.__write_pipe)
                 os._exit(status) # pylint: disable=protected-access
@@ -809,18 +916,42 @@ class ConfigFunctionFFSSubprocess(ConfigFunctionFFS):
             if exc.errno != errno.ESRCH:
                 raise
 
+    def _updateExitStatus(self, blocking):
+        if self.__exit_status is not None:
+            return
+        try:
+            wait_id_result = os.waitid(
+                os.P_PID,
+                self.__pid,
+                os.WEXITED | (
+                    0
+                    if blocking else
+                    os.WNOHANG
+                ),
+            )
+        except OSError as exc:
+            if exc.errno != errno.ECHILD:
+                raise
+            # Uh oh, child process exited and we did not get to read its exit
+            # status. Pick the generic error status.
+            # XXX: warn user ?
+            self.__exit_status = 1
+        else:
+            if wait_id_result is not None:
+                self.__exit_status = (
+                    wait_id_result.si_status
+                    if wait_id_result.si_code == os.CLD_EXITED else
+                    # Subprocess was killed or dumped core, pick the generic error
+                    # status.
+                    1
+                )
+
     def join(self):
         """
         Wait for function subprocess to exit.
         """
-        try:
-            os.waitpid(self.__pid, 0)
-        except OSError as exc:
-            # If the child process does not exist (no status to reap), then
-            # all good.
-            if exc.errno != errno.ECHILD:
-                raise
-        super().join()
+        self._updateExitStatus(blocking=True)
+        self._umount()
 
     def run(self):
         """
@@ -833,3 +964,12 @@ class ConfigFunctionFFSSubprocess(ConfigFunctionFFS):
             self.function.processEventsForever()
         except KeyboardInterrupt:
             pass
+
+    def getExitStatus(self):
+        """
+        Return the subprocess' exit status.
+
+        For use in the gadget process to inspect the function's exit status.
+        """
+        self._updateExitStatus(blocking=False)
+        return self.__exit_status
